package org.openjdk.varhandles;


import freemarker.cache.ClassTemplateLoader;
import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;

import java.io.*;
import java.util.*;

public class Main {

    public static final Map<String, String> VALS;
    public static final Map<String, String> VALS_ALT;

    static {
        VALS = new HashMap<>();
        VALS.put("boolean", "true");
        VALS.put("byte",    "42");
        VALS.put("short",   "42");
        VALS.put("char",    "'a'");
        VALS.put("int",     "42");
        VALS.put("float",   "42f");
        VALS.put("long",    "42L");
        VALS.put("double",  "42d");
        VALS.put("Object",  "new Object()");

        VALS_ALT = new HashMap<>();
        VALS_ALT.put("boolean", "false");
        VALS_ALT.put("byte",    "43");
        VALS_ALT.put("short",   "43");
        VALS_ALT.put("char",    "'b'");
        VALS_ALT.put("int",     "43");
        VALS_ALT.put("float",   "43f");
        VALS_ALT.put("long",    "43L");
        VALS_ALT.put("double",  "43d");
        VALS_ALT.put("Object",  "new Object()");
    }

    public static void main(String... args) {
        if (args.length == 0) {
            throw new IllegalStateException("Need an argument");
        }

        String target = args[0];

        Set<String> ALL_TYPES = new HashSet<>();
        ALL_TYPES.add("boolean");
        ALL_TYPES.add("byte");
        ALL_TYPES.add("short");
        ALL_TYPES.add("char");
        ALL_TYPES.add("int");
        ALL_TYPES.add("float");
        ALL_TYPES.add("long");
        ALL_TYPES.add("double");
        ALL_TYPES.add("Object");

        Set<String> ALL_TESTS = new HashSet<>();
        ALL_TESTS.add("getVolatile");
        ALL_TESTS.add("setVolatile");
        ALL_TESTS.add("getAcquire");
        ALL_TESTS.add("getAcquireStatic");
        ALL_TESTS.add("setRelease");
        ALL_TESTS.add("getRelaxed");
        ALL_TESTS.add("setRelaxed");
        ALL_TESTS.add("getOpaque");
        ALL_TESTS.add("setOpaque");

        generate(target, ALL_TESTS, ALL_TYPES);

        Set<String> CAS_TESTS = new HashSet<>();
        CAS_TESTS.add("cas");
        CAS_TESTS.add("caeAcquire");
        CAS_TESTS.add("caeRelease");
        CAS_TESTS.add("caeVolatile");
        CAS_TESTS.add("weakCas");
        CAS_TESTS.add("weakCasAcquire");
        CAS_TESTS.add("weakCasRelease");

        Set<String> CAS_TYPES = new HashSet<>();
        CAS_TYPES.add("int");
        CAS_TYPES.add("long");
        CAS_TYPES.add("Object");

        generate(target, CAS_TESTS, CAS_TYPES);
    }

    private static void generate(String target, Set<String> tests, Set<String> types) {
        //Freemarker configuration object
        Configuration cfg = new Configuration();
        try {
            //Load template from source folder
            cfg.setTemplateLoader(new ClassTemplateLoader(ClassLoader.getSystemClassLoader(), ""));

            for (String test : tests) {
                Template template = cfg.getTemplate("/" + test + "-test.template");

                for (String type : types) {
                    String Type = Character.toUpperCase(type.charAt(0)) + type.substring(1);
                    String testName = Type + "Test";
                    String packageName = "org.openjdk.varhandles." + test;

                    Map<String, Object> data = new HashMap<>();
                    data.put("testName", testName);
                    data.put("package", packageName);
                    data.put("val", VALS.get(type));
                    data.put("valAlt", VALS_ALT.get(type));
                    data.put("type", type);
                    data.put("Type", Type);

                    String className = packageName + "." + testName;

                    File d = new File(target + "/" + packageName.replaceAll("\\.", File.separator));
                    d.mkdirs();

                    File f = new File(target + "/" + className.replaceAll("\\.", File.separator) + ".java");

                    // Console output
                    Writer out = new PrintWriter(f);
                    template.process(data, out);
                    out.flush();
                    out.close();
                }
            }

        } catch (IOException | TemplateException e) {
            e.printStackTrace();
        }
    }

}
